from pmlab_lite.log import Event, EventLog
from . import constants as  c

from dateutil.parser import parse #this is slow compared to the following two
import ciso8601
from datetime import datetime # datetime.fromisoformat is fast, but only implemented from >=3.7
import re
import gzip
from lxml import etree
from tqdm import tqdm
import logging

__START = c.START
__END = c.END
__actions = (__START, __END)

"""
    Note:
    The code for importing (__parse_xes_file, __parse_attribute) is inspired and partly taken from the open source process minind library pm4py (https://pm4py.fit.fraunhofer.de).
    Thanks to all involved researchers / developers, especially Alessandro Berti and Sebastiaan van Zelst, for providing the well documented
    and written code and providing this as an open source project.
"""


def import_xes(file):
    """
        Reads a *.xes-file and returns it as a log data structure.

        Args:
            file: string literal path to a *.xes-file

        Returns:
            log: the log represented as a data structure as in pmlab_lite.log
    """

    print("Processing log...")

    file = __check_file_type(file)
    f = open(file, "rb")

    context = etree.iterparse(f, events=__actions)
    tree = {} #for keeping only what we need to know (we delete elements in iterparse iterator to keep memory usage minimal)
    log = __parse_xes_file(context, tree)

    del context
    f.close()

    return log

def __check_file_type(file) -> str:
    """
        Checks for possiblities of files to input in the importer.

        Returns the needed filename as a string.
    """

    if isinstance(file, str):
        filename = file
        if filename.endswith('.gz'):
            file = gzip.open(filename, 'rb')
        else:
            pass  # Just send the filename to xmltree.parse
    else:
        filename = file.name

    return filename

def __clear_element(element, tree):
    """ Deletes an element from tree, to avoid memory explosion, when not needed anymore"""
    if element in tree:
      del tree[element]
    element.clear()
    if element.getprevious() is not None:
        try:
            del element.getparent()[0]
        except TypeError:
            pass

def __parse_date_ciso(date_str: str):
    """ Parses a date string from an xes log file to a datetime object. Uses external ciso library."""

    return ciso8601.parse_datetime(date_str)

def __parse_attribute(elem, store, key, value, tree):
  #store is basically the parent and can be:

  if len(elem.getchildren()) == 0:  #elem is leaf node, i.e. no further

    if type(store) is list:
      # changes to the store of lists: not dictionaries anymore
      # but pairs of key-values.
      store.append((key, value))
    else:
      store[key] = value

  else:
    if elem.getchildren()[0].tag.endswith(c.VLS):
      store[key] = {c.VAL: value, c.CLN: list()}
      tree[elem] = store[key][c.CLN]
      tree[elem.getchildren()[0]] = tree[elem]
    else:
      store[key] = {c.VAL: value, c.CLN: dict()}
      tree[elem] = store[key][c.CLN]

  return tree

def __parse_xes_file(context, tree):

    log = None
    trace_id = None
    event = None
    trace_count = 0 #for trace_id

    for action, elem in tqdm(context):

        if action == __START:  # starting to read

            parent = tree[elem.getparent()] if elem.getparent() in tree else None

            if elem.tag.endswith(c.STR):
              if parent is not None:
                tree = __parse_attribute(elem, parent, elem.get(c.KEY), elem.get(c.VAL), tree)
              continue

            elif elem.tag.endswith(c.DTE):
              try:
                date = __parse_date_ciso(elem.get(c.VAL))
                tree = __parse_attribute(elem, parent, elem.get(c.KEY), date, tree)
              except TypeError:
                logging.info("failed to parse date: " + str(elem.get(c.VAL)))
              except ValueError:
                logging.info("failed to parse date: " + str(elem.get(c.VAL)))
              continue

            elif elem.tag.endswith(c.EVE):
              if event is not None:
                raise SyntaxError('file contains <event> in another <event> tag')
              if trace_id is None:
                raise SyntaxError('file contains a <event> element outside of a <trace> element (trace_id is None)')
              event = Event(trace_id)
              tree[elem] = event
              continue

            elif elem.tag.endswith(c.TRC):
              if trace_id is not None:
                raise SyntaxError('file contains <trace> in another <trace> tag')
              trace_count += 1
              trace_id = trace_count
              log.add_trace(trace_id)
              tree[elem] = log.traces[trace_id]
              continue

            # rest of elif's belong here

            elif elem.tag.endswith(c.EXT):
              if log is None:
                raise SyntaxError('extension found outside of <log> tag')
              if elem.get(c.NME) is not None and elem.get(c.PRX) is not None and elem.get(c.URI) is not None:
                log.extensions[elem.attrib[c.NME]] = {c.PRX: elem.attrib[c.PRX], c.URI: elem.attrib[c.URI]}
              continue

            elif elem.tag.endswith(c.GLB):
              if log is None:
                raise SyntaxError('global found outside of <log> tag')
              if elem.get(c.SCP) is not None:
                log.globals[elem.attrib[c.SCP]] = {}
                tree[elem] = log.globals[elem.get(c.SCP)]
              continue

            elif elem.tag.endswith(c.CLS):
              if log is None:
                raise SyntaxError('classifier found outside of <log> tag')
              if elem.get(c.KYS) is not None:
                name, attributes  = elem.attrib[c.NME], elem.attrib[c.KYS]
                # resolve special case; refactor with __parse_classifier
                if "'" in attributes:
                  attributes = re.findall(r'\'(.*?)\'', attributes)
                else:
                  attributes = attributes.split()
                log.classifiers[name] = attributes
              continue

            elif elem.tag.endswith(c.LOG):
              if log is not None:
                raise SyntaxError('file contains > 1 <log> tags')
              log = EventLog()
              tree[elem] = log.attributes
              continue

        elif action == __END:

            __clear_element(elem, tree)

            if elem.tag.endswith(c.EVE):
              if trace_id is not None: # here originally was written "if trace is not None", we dont use traces as objects though
                log.add_event(event)
                event = None
              continue

            elif elem.tag.endswith(c.TRC):
              trace_id = None
              continue

            elif elem.tag.endswith(c.LOG):
              continue

    return log



def export_to_xes(log: EventLog, target_path: str):
    """Exprots an EventLog structure as stored by pmlab_lite to an *.xes-file.

    Args:
        log (EventLog): The Event Log to be exported.
        target_path (str): The path- and filename of the exported *.xes-file, e.g. "some_file.xes".
    """

    root = etree.Element('log')

    __export_extensions(log, root)
    __export_globals(log, root)
    __export_classifiers(log, root)
    __export_log_attributes(log, root)
    __export_traces(log, root)

    tree = etree.ElementTree(root)
    tree.write(target_path, pretty_print=True, xml_declaration=True, encoding="utf-8")

def __export_extensions(log: EventLog, root):
    for key in log.extensions:
        extension = etree.SubElement(root, 'extension')
        extension.set('name', key)
        extension.set('prefix', log.extensions[key]['prefix'])
        extension.set('uri', log.extensions[key]['uri'])

def __export_globals(log: EventLog, root):
    for scope in log.globals:
        g = etree.SubElement(root, 'global')
        g.set('scope', scope)
        __export_attributes(log.globals[scope], g)

def __export_classifiers(log: EventLog, root):
    for classi in log.classifiers.keys():
        classi_attributes = ['\'' + attr + '\'' if ' ' in attr else attr for attr in log.classifiers[classi]]
        classifier = etree.SubElement(root, 'classifier')

        classifier.set('name', classi)
        classifier.set('keys', " ".join(classi_attributes))

def __export_log_attributes(log: EventLog, root):
    __export_attributes(log.attributes, root)

def __export_traces(log: EventLog, root):
    for case_id in log.traces.keys():
        trace = etree.SubElement(root, 'trace')
        trace_attr = {key:log.traces[case_id][key] for key in log.traces[case_id] if key!='events'} #don't consider the events as a trace attribute, alternate solution: make all atributes a subdict
        __export_attributes(trace_attr, trace)
        __export_events_of_trace(log.traces[case_id]['events'], trace)

def __export_events_of_trace(trc: list, root):
    for evnt in trc:
        event = etree.SubElement(root, 'event')
        __export_attributes(evnt, event)

def __export_attributes(attributes: dict, root):
    for log_attr, log_attr_val in attributes.items():
        log_attr_type = type(log_attr_val).__name__
        xes_attr_type = __log_to_xes_attribute_type(log_attr_type)
        xes_attr_value = __get_xes_attribute_value(xes_attr_type, log_attr_val)

        attribute = etree.SubElement(root, xes_attr_type)
        attribute.set('key', log_attr)
        attribute.set('value', xes_attr_value)

def __log_to_xes_attribute_type(log_attr_type) -> str:
    if log_attr_type in c.xes_typenames:
        return c.xes_typenames[log_attr_type]
    else:
        return 'string'

def __get_xes_attribute_value(xes_attr_type, log_attr_val):
    if xes_attr_type == 'date':
        return log_attr_val.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + log_attr_val.strftime('%z')[0:3] + ':' + log_attr_val.strftime('%z')[3:]
    elif xes_attr_type == 'boolean':
        return str(log_attr_val).lower()
    else:
        return str(log_attr_val)



import unittest
from pmlab_lite.helper.io import xes
from pprint import pprint

a_names = ['FlagX1.xes', 'FlagX2.xes', 'LevelA1.xes', 'LevelA2.xes', 'LevelB1.xes', 'LevelB2.xes',
           'LevelC1.xes', 'LevelC2.xes', 'LevelD1.xes', 'LevelD2.xes']

rl_names = ['BPIC12.xes', 'BPIC13_closed_problems.xes', 'BPIC13_incidents.xes', 'BPIC13_open_problems.xes',
            'BPIC15_1.xes', 'BPIC15_2.xes', 'BPIC15_3.xes', 'BPIC15_4.xes', 'BPIC15_5.xes', 'BPIC17.xes', 'BPIC17_Offer_log.xes']

a_path = '../XES_Certification/Artificial/'
rl_path = '../XES_Certification/Real-life/'

a_log = a_path + a_names[1]
rl_log = rl_path + rl_names[9]


# old log
#log = xes.import_from_xes(a_log)

# new log
new_log = xes.import_xes(a_log)

# pm4py log
from pm4py.objects.log.importer.xes import importer as xes_importer

#test_log = xes_importer.apply(rl_log, variant=xes_importer.Variants.ITERPARSE)

# compare extensions --> success
""" print()
pprint(log.extensions)
print()
pprint(new_log.extensions)
print("Extensions are the same:", log.extensions==new_log.extensions)
print()
pprint(test_log.extensions)
print() """

# compare globals --> success
""" print()
pprint(log.globals)
print()
pprint(new_log.globals)
print("Globals are the same:", log.globals==new_log.globals)
print()
pprint(test_log.omni_present)
print() """

# compare classifiers --> success
""" print()
pprint(log.classifiers)
print()
pprint(new_log.classifiers)
print("Classifiers are the same:", log.classifiers==new_log.classifiers)
print()
pprint(test_log.classifiers)
print() """


# compare log attributes --> success
""" print()
pprint(log.attributes)
print()
pprint(new_log.attributes)
print("Log attributes are the same:", log.attributes==new_log.attributes)
print()
pprint(test_log.attributes)
print() """

# compare traces --> succes but former case_id, which was equivalent to the concept:name of a trace, is no a count from 0 to num_traces
""" traces = log.traces
new_traces = new_log.traces
for i, trace_id in enumerate(traces):

    new_traces[trace_id] = new_traces.pop(i+1)
    new_traces[trace_id]['concept:name'] = trace_id

    for event in new_traces[trace_id]['events']:
        event['case:id'] = trace_id

    if i == 1:
        pprint(traces[trace_id])
        print()
        pprint(new_traces[trace_id])

print("Events of traces are equal:", traces==new_traces, "(besides case:id)") """

#compare events --> success
""" events = log.events
pprint(log.events[:3])
print()
new_events = new_log.events
pprint(new_log.events[:3])

for event in events:
    del event['case:id']
for event in new_events:
    del event['case:id']

print("Events are the are equal:", events == new_events, "(besides case_id)") """


# test exporting
#xes.export_to_xes(new_log, 'X2.xes')



### xes element constants
EVE = 'event'
TRC = 'trace'
LOG = 'log'
EXT = 'extension'
GLB = 'global'
SCP = 'scope'
CLS = 'classifier'
PRX = 'prefix'
URI = 'uri'
NME = 'name'
KEY = 'key'
VAL = 'value'
KYS = 'keys'
VLS = 'values'
CLN = 'children'

STR = 'string'
DTE = 'date'

### holds the equivalent xes type names for the python type names
xes_typenames = {"str": 'string',
                 "datetime": 'date',
                 "Timestamp": 'date',
                 "int": 'int',
                 "float": 'float',
                 "bool": 'boolean',
                 "dict": 'list'
                }

NAMESPACES = {'xes': 'http://www.xes-standard.org/'}
ATTR_TYPES = ['string', 'date', 'boolean', 'int']

### lxml iterparse constants
START = 'start'
END = 'end'


import unittest
from pmlab_lite.pn import PetriNet, TraceNet, SynchronousProduct
from pmlab_lite.alignments import a_star
from pmlab_lite.helper.viz.dot import draw_petri_net, draw_synchronous_product, draw_a_star_search_space
import numpy as np
from numpy.linalg import lstsq, matrix_rank
from cvxopt import matrix, solvers

net = PetriNet()
transitions = ['A', 'B']
edges = [(1,-1), (-1,2), (2,-2), (-2,3)]
for p in range(1,4):
    net.add_place(p)
for t in transitions:
    net.add_transition(t)
for e in edges:
    net.add_edge(e[0], e[1])

trace = ['A', 'B']
trace_net = TraceNet(trace)

sp = SynchronousProduct(net, trace_net)
im = sp.incidence_matrix()
imv = sp.get_init_marking()
fmv = sp.get_final_marking()

b = fmv - imv
x = np.linalg.lstsq(im, b, rcond=None)[0]
#print(x)
#x[x > 0] = 1
#x[x <= 0] = 0
trans_to_idx = sp.transitions_by_index()
for key in trans_to_idx:
    if trans_to_idx[key].startswith('tau') or trans_to_idx[key].endswith('synchronous'):
        x[key] = 0

# print(np.argmax(x))
# print(np.argwhere(x == x.max()).flatten().tolist())

#draw_petri_net(net, filename='../lp_net')
#draw_petri_net(trace_net, filename='../trace_net')
#draw_synchronous_product(synchronous_product, filename='../synchronous_product')

print("Incidence matrix:")
print(im)
print("Final marking  :",fmv)
print("Current marking:",imv)
print("Solution       :",x)
print("Solutions cost :",x.sum())
print("Optimal cost would be:", 0.0)
print("--> overestimating the total cost:", x.sum()>0.0)
print()

# computing solution when all "redundant" columns in the incedence marix are left out
# here this means transition0 and transition2 are covered by transition 4 (similar for 1,3 covered by 5)
im2 = im[:, 4:]
x2 = np.linalg.lstsq(im2, b, rcond=None)[0]
print("Incidence matrix:")
print(im2)
print("Solution       :",x2)
print("Solutions cost :",x2.sum()*0.0) # pretending to have remembered which index was a synchronous move in the new inc-mat and thus aranging costs
print("Optimal cost would be:", 0.0)
print("--> overestimating the total cost:", False)

# formulate as linear program for cvxopt
# N = sp.num_transitions()
# d = net.num_transitions() + trace_net.num_transitions()
#
# # minimization term
# c = np.zeros((N,1))
# #c[[idx for idx in range(d)]] = 1.0
# c = matrix(c)
#
# # inequality constraints
# G = matrix(-np.eye(N))
# h = matrix(np.zeros(N))
#
# # equality constraints
# A = matrix(im[:,:4]*1.0)
# b = matrix((fmv-imv)*1.0)
#print(A)
#print("Rank of A:", matrix_rank(A), "\n And number of rows:", A.size[0])
#sol = solvers.lp(c, G, h, A, b)

# formulate as an cvxopt quadratic problem
# N = sp.num_transitions()
# d = net.num_transitions() + trace_net.num_transitions()
#
#
# P = matrix(np.zeros((N,N)))
#
# q = np.zeros((N,1))
# q[ [idx for idx in range(d)] ] = 1.0
# q = matrix(q)
#
# G = matrix(-np.eye(N))
# h = matrix(np.zeros(N))
# A = matrix((im*1.0))
# b = matrix((fmv - imv)*1.0)
#
# sol = solvers.qp(P, q, G, h, A, b)
#
# print(P)
# print(q)
# print(G)
# print(h)
# print(A)
# print(b)
# print("CVXOPT quadratic programm solution:",sol['x'])
